# $Id: component.rb,v 1.3 2005/08/22 09:12:03 aamine Exp $

require 'rational'
require 'forwardable'

class IntegerOverflow < ::StandardError; end
class ParseError < ::StandardError; end

class Component
  @LLDN = nil
  def Component.LLDN
    @LLDN ||= defcomponentlist(
      %w(       ϻ  Ȭ  ),
      ['̵',    8],
      ['ԲĻ׵',    8],
      ['ͳ¿',      8],
      ['ε',      8],
      ['Ϻ',      8],
      ['',          4],
      ['',          4],
      ['',          4],
      ['',          4],
      ['',          4],
      ['',          4],
      ['ͽ',        4],
      ['Զ',          4],
      ['',          4],
      ['',          4],
      ['',          4],
      ['',          1],
      ['',          1],
      ['ɴ',          1],
      ['',          1],
      [nil,           0]
    )
  end

  def Component.defcomponentlist(nums, *deflist)
    term = Terminator.new(nums)
    list = deflist.inject(nil) {|_next, a| new(_next, *a) }
    list.each do |c|
      c.instance_eval {
        pow = next_pow()
        @branch = (pow > 1 ? list.take(pow) : term)
      }
    end
    Head.new(list)
  end

  include Enumerable

  def initialize(_next, unit, pow)
    @next = _next
    @unit = unit
    @pow = pow
  end

  attr_reader :unit
  attr_reader :pow
  attr_reader :branch

  def inspect
    "#<Component #{@unit.to_s}>"
  end

  def next_pow
    @next ? @next.pow : 8
  end

  def each(&block)
    yield self
    @next.each(&block) if @next
  end

  def length
    inject(0) {|len, i| len + 1 }
  end

  def take(n)
    if n - @pow > 0
    then dup(@next.take(n - @pow))
    else nil
    end
  end

  def dup(_next = @next, branch = @branch)
    obj = self.class.new(_next, @unit, @pow)
    obj.instance_eval { @branch = branch }
    obj
  end

  def chars_re
    @chars_re ||= /#{Regexp.union(*chars())}+/
  end

  def chars
    inject([]) {|set, c| set + [c.chars1] }.flatten.uniq.compact
  end

  def chars1
    @branch.inject([@unit]) {|set, c| set + c.chars1 }
  end

  def parse(str)
    if m = regexp().match(str)
    then upnum(m.pre_match) + @branch.parse(m[1])
    else upnum(str)
    end
  end

  def upnum(s)
    return 0 if s.empty?
    raise IntegerOverflow, "integer too big" unless @next
    @next.parse(s) * (10 ** @next.pow)
  end
  private :upnum

  def regexp
    @re ||= /(#{@branch.pattern})#{@unit ? '?' : ''}#{@unit.to_s}\z/
  end
  private :regexp

  def pattern
    (@next ? @next.pattern : '') + "(?:(#{@branch.pattern})?#{@unit.to_s})?"
  end

  def string_expr(n)
    up, base = *n.divmod(10 ** next_pow())
    upstr(up) + basestr(base)
  end

  def upstr(n)
    unless @next
      return '' if n == 0
      raise IntegerOverflow, "integer too big"
    end
    @next.string_expr(n)
  end
  private :upstr

  def basestr(n)
    if n == 0
    then ''
    else "#{@branch.string_expr(n)}#{@unit}"
    end
  end
  private :basestr
end

class Terminator
  include Enumerable

  def initialize(nums)
    @nums = nums
    @re = nil
  end

  def inspect
    "#<TERM>"
  end

  def each
    yield self
  end

  def chars1
    @nums
  end

  def pattern
    "[#{@nums.join()}]"
  end

  def parse(str)
    return Rational(1) unless str
    return Rational(1) if str.empty?
    n = @nums.index(str) or
        raise ParseError, "unexpected string: #{str.inspect}"
    Rational(n)
  end

  def string_expr(n)
    @nums[n]
  end
end

class Head
  def initialize(list)
    @list = list
  end

  def inspect
    "#<HEAD>"
  end

  extend Forwardable
  def_delegator "@list", :chars_re
  def_delegator "@list", :parse
  
  def string_expr(n)
    str = @list.string_expr(n.to_i.abs).gsub(/(?=[ɴ])/, '')
    str = '' if str.empty?
    (n.to_i < 0 ? '' : '') + str + (n.denominator==1 ? '' : ';')
  end
end
